'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _ = _interopDefault(require('lodash'));
var humanize = _interopDefault(require('string-humanize'));
var retrieveArguments = _interopDefault(require('retrieve-arguments'));
var fs = _interopDefault(require('fs'));
var crypto = _interopDefault(require('crypto'));

var asyncToGenerator = function (fn) {
  return function () {
    var gen = fn.apply(this, arguments);
    return new Promise(function (resolve, reject) {
      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          return Promise.resolve(value).then(function (value) {
            step("next", value);
          }, function (err) {
            step("throw", err);
          });
        }
      }

      return step("next");
    });
  };
};

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();







var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};



var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};











var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};





var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();











var toArray = function (arr) {
  return Array.isArray(arr) ? arr : Array.from(arr);
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var Stage = function () {
    function Stage() {
        classCallCheck(this, Stage);
    }

    createClass(Stage, [{
        key: "and",
        value: function and() {
            return this;
        }
    }, {
        key: "but",
        value: function but() {
            return this;
        }
    }, {
        key: "with",
        value: function _with() {
            return this;
        }
    }, {
        key: "given",
        value: function given() {
            return this;
        }
    }, {
        key: "when",
        value: function when() {
            return this;
        }
    }, {
        key: "then",
        value: function then() {
            return this;
        }
    }]);
    return Stage;
}();

function getStageMetadataStoreProvider(key) {
    var KEY = `__JSGIVEN__INTERNAL__${key}__KEY__`;

    function getOrBuildStore(target) {
        if (!target[KEY]) {
            target[KEY] = new StageMetadataStoreImpl();
        }

        return target[KEY];
    }

    return {
        getStoreFromTarget(target) {
            return getOrBuildStore(target);
        },
        getStoreFromStageClass(stageClass) {
            return getOrBuildStore(stageClass.prototype);
        }
    };
}

var StageMetadataStoreImpl = function () {
    function StageMetadataStoreImpl() {
        classCallCheck(this, StageMetadataStoreImpl);
        this.properties = [];
    }

    createClass(StageMetadataStoreImpl, [{
        key: 'addProperty',
        value: function addProperty(property) {
            this.properties.push(property);
        }
    }, {
        key: 'getProperties',
        value: function getProperties() {
            return this.properties;
        }
    }]);
    return StageMetadataStoreImpl;
}();

function checkIsFunction(target, errorMessage) {
    if (!target || !_.isFunction(target)) {
        throw new Error(errorMessage);
    }
}

function checkIsParameter(func, parameterName, errorMessage) {
    if (!retrieveArguments(func).includes(parameterName)) {
        throw new Error(errorMessage);
    }
}

var storeProvider = getStageMetadataStoreProvider('@ParameterFormatters');

function getFormatters(stage, requestedStepMethodName, requestedParameterName) {
    var formattings = storeProvider.getStoreFromTarget(stage).getProperties();
    return formattings.filter(function (_ref) {
        var stepMethodName = _ref.stepMethodName,
            parameterName = _ref.parameterName;
        return requestedStepMethodName === stepMethodName && requestedParameterName === parameterName;
    }).map(function (_ref2) {
        var formatter = _ref2.formatter;
        return formatter;
    });
}

function restParameterName() {
    return 'JSGIVEN_REST_PARAMETER_NAME';
}

function buildParameterFormatter(formatter) {
    var parameterFormatter = function parameterFormatter() {
        for (var _len = arguments.length, parameterNames = Array(_len), _key = 0; _key < _len; _key++) {
            parameterNames[_key] = arguments[_key];
        }

        var decorator = function decorator(target, stepMethodName, descriptor) {
            parameterNames.forEach(function (parameterName) {
                checkIsFunction(target[stepMethodName], `Formatter decorators can only be applied to methods: '${stepMethodName}' is not a method.`);
                checkIsParameter(target[stepMethodName], parameterName, `Formatter decorator cannot be applied on method: ${stepMethodName}(): parameter '${parameterName}' was not found.`);
                storeProvider.getStoreFromTarget(target).addProperty({
                    formatter,
                    parameterName,
                    stepMethodName
                });
            });
            return _extends({}, descriptor, { writable: true });
        };
        return decorator;
    };
    parameterFormatter.formatParameter = function (stageClass, stepMethodName) {
        for (var _len2 = arguments.length, parameterNames = Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
            parameterNames[_key2 - 2] = arguments[_key2];
        }

        parameterNames.forEach(function (parameterName) {
            checkIsFunction(
            // $FlowIgnore
            stageClass.prototype[stepMethodName], `Formatter.formatParameter() can only be applied to methods: '${stepMethodName}' is not a method.`);
            checkIsParameter(
            // $FlowIgnore
            stageClass.prototype[stepMethodName], parameterName, `Formatter.formatParameter() cannot be applied on method: ${stepMethodName}(): parameter '${parameterName}' was not found.`);
            storeProvider.getStoreFromStageClass(stageClass).addProperty({
                formatter,
                parameterName,
                stepMethodName
            });
        });
    };

    return parameterFormatter;
}

var Quoted = buildParameterFormatter(function (parameterValue) {
    return `"${parameterValue}"`;
});

var QuotedWith = function QuotedWith(quoteCharacter) {
    return buildParameterFormatter(function (parameterValue) {
        return `${quoteCharacter}${parameterValue}${quoteCharacter}`;
    });
};

var NotFormatter = buildParameterFormatter(function (parameterValue) {
    return parameterValue ? '' : 'not';
});

var INTRO_WORD_METHODS = ['given', 'when', 'then', 'and', 'but', 'with'];

var ScenarioPart = function () {
    function ScenarioPart(kind) {
        var steps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
        classCallCheck(this, ScenarioPart);

        this.kind = kind;
        this.steps = steps;
        this.introWord = null;
    }

    createClass(ScenarioPart, [{
        key: 'stageMethodCalled',
        value: function stageMethodCalled(methodName, parameters) {
            if (INTRO_WORD_METHODS.find(function (introWord) {
                return introWord === methodName;
            })) {
                this.introWord = methodName;
            } else {
                var isFirstStep = this.steps.length === 0;
                this.steps.push(new Step(methodName, parameters, isFirstStep, this.introWord));
                this.introWord = null;
            }
        }
    }]);
    return ScenarioPart;
}();

var Step = function Step(methodName, parameters, isFirstStep, introWord) {
    classCallCheck(this, Step);

    var TWO_DOLLAR_PLACEHOLDER = 'zzblablaescapedollarsignplaceholdertpolm';

    var parametersCopy = [].concat(toConsumableArray(parameters));

    var words = [].concat(toConsumableArray(methodName // 'a_bill_of_$_$$'
    .replace('$$', TWO_DOLLAR_PLACEHOLDER // 'a_bill_of_$_TWO_DOLLAR_PLACEHOLDER'
    ).split('$' // ['a_bill_of', 'TWO_DOLLAR_PLACEHOLDER']
    ).map(function (word) {
        return _.lowerCase(humanize(word));
    } //  ['a bill of', 'TWO_DOLLAR_PLACEHOLDER']
    ).reduce(function (previous, newString, index) {
        if (index === 0) {
            return [{ word: newString, scenarioParameterName: null }];
        }

        var formattedParameters = void 0;
        if (parametersCopy.length > 0) {
            var _parametersCopy$splic = parametersCopy.splice(0, 1),
                _parametersCopy$splic2 = slicedToArray(_parametersCopy$splic, 1),
                parameter = _parametersCopy$splic2[0];

            var _word = formatParameter(parameter.value, parameter.formatters);
            formattedParameters = [{
                word: _word,
                scenarioParameterName: parameter.scenarioParameterName
            }];
        } else {
            formattedParameters = [];
        }

        return [].concat(toConsumableArray(previous), toConsumableArray(formattedParameters), [{ word: newString, scenarioParameterName: null }]);
    }, [] //  ['a bill of', '500', 'TWO_DOLLAR_PLACEHOLDER']
    ).filter(function (_ref) {
        var word = _ref.word;
        return word !== '';
    } // If one puts a $ at the end of the method, this adds a useless '' at the end
    ).map(function (_ref2) {
        var word = _ref2.word,
            scenarioParameterName = _ref2.scenarioParameterName;
        return {
            word: word.replace(TWO_DOLLAR_PLACEHOLDER, '$'),
            scenarioParameterName
        };
    } //  ['a bill of', '500', '$']
    ).map(toWord)), toConsumableArray(parametersCopy.map(function (parameter) {
        return {
            word: formatParameter(parameter.value, parameter.formatters),
            scenarioParameterName: parameter.scenarioParameterName
        };
    }).map(toWord)));

    if (introWord) {
        words = [toIntroWord(introWord)].concat(toConsumableArray(words));
    }

    if (isFirstStep) {
        var _words = words,
            _words2 = toArray(_words),
            _words2$ = _words2[0],
            value = _words2$.value,
            isIntroWord = _words2$.isIntroWord,
            rest = _words2.slice(1);

        words = [new Word(_.upperFirst(value), isIntroWord)].concat(toConsumableArray(rest));
    }
    this.words = words;
    this.name = words.map(function (_ref3) {
        var value = _ref3.value;
        return value;
    }).join(' ');

    function toWord(_ref4) {
        var word = _ref4.word,
            scenarioParameterName = _ref4.scenarioParameterName;

        return new Word(word, false, scenarioParameterName);
    }

    function toIntroWord(value) {
        return new Word(value, true, null);
    }
};

var Word = function Word(value, isIntroWord) {
    var scenarioParameterName = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    classCallCheck(this, Word);

    this.value = value;
    this.isIntroWord = isIntroWord;
    this.scenarioParameterName = scenarioParameterName;
};

var ScenarioCase = function ScenarioCase(args) {
    var parts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    classCallCheck(this, ScenarioCase);

    this.args = args;
    this.parts = parts;
};

var ScenarioReport = function () {
    function ScenarioReport(groupReport, name, cases, argumentNames) {
        classCallCheck(this, ScenarioReport);

        this.groupReport = groupReport;
        this.name = name;
        this.cases = cases;
        this.argumentNames = argumentNames;
    }

    createClass(ScenarioReport, [{
        key: 'dumpToFile',
        value: function dumpToFile(reportsDestination) {
            createDirOrDoNothingIfExists(reportsDestination);
            var fileName = computeScenarioFileName(this.groupReport.name, this.name);
            fs.writeFileSync(`${reportsDestination}/${fileName}`, JSON.stringify(this), 'utf-8');
        }
    }]);
    return ScenarioReport;
}();

function formatParameter(parameter, formatters) {
    if (formatters.length > 0) {
        var value = parameter;
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = formatters[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var formatter = _step.value;

                value = formatter(value);
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        return formatParameter(value, []);
    }
    if (_.isString(parameter)) {
        return parameter;
    } else if (_.isObject(parameter) || Array.isArray(parameter)) {
        if (parameter.toString && parameter.toString !== Object.prototype.toString && parameter.toString !== Array.prototype.toString) {
            return parameter.toString();
        }
        return JSON.stringify(parameter);
    } else {
        return parameter && parameter.toString ? parameter.toString() : JSON.stringify(parameter);
    }
}

var GroupReport = function GroupReport(name) {
    classCallCheck(this, GroupReport);

    this.name = name;
};

function createDirOrDoNothingIfExists(path) {
    try {
        fs.mkdirSync(path);
    } catch (error) {
        if (error.code !== 'EEXIST') {
            throw error;
        } else {
            // do nothing
        }
    }
}

function computeScenarioFileName(groupName, scenarioName) {
    var hash = crypto.createHash('sha256');
    hash.update(groupName + '\n' + scenarioName);
    return hash.digest('hex');
}

var stateProvider = getStageMetadataStoreProvider('@State');

function State(target, key, descriptor) {
    stateProvider.getStoreFromTarget(target).addProperty(key);
    return _extends({}, descriptor, { writable: true });
}
State.addProperty = function (stageClass, property) {
    stateProvider.getStoreFromStageClass(stageClass).addProperty(property);
};

function copyStateProperties(source, target) {
    if (source && target) {
        var propertyNames = _.intersection(stateProvider.getStoreFromTarget(source).getProperties(), stateProvider.getStoreFromTarget(target).getProperties());
        propertyNames.forEach(function (propertyName) {
            // Need to convert to any to avoid typechecking
            var sourceAny = source;
            var targetAny = target;
            targetAny[propertyName] = sourceAny[propertyName];
        });
    }
}

var storeProvider$1 = getStageMetadataStoreProvider('@Hidden');

function Hidden(target, key, descriptor) {
    checkIsFunction(target[key], `@Hidden decorator can only be applied to methods: '${key}' is not a method.`);
    storeProvider$1.getStoreFromTarget(target).addProperty(key);
    return _extends({}, descriptor, { writable: true });
}
Hidden.addHiddenStep = function (stageClass, property) {
    checkIsFunction(
    // $FlowIgnore
    stageClass.prototype[property], `Hidden.addHiddenStep() can only be applied to methods: '${property}' is not a method.`);
    storeProvider$1.getStoreFromStageClass(stageClass).addProperty(property);
};

function isHiddenStep(stage, stepName) {
    var steps = storeProvider$1.getStoreFromTarget(stage).getProperties();
    return steps.includes(stepName);
}

var REPORTS_DESTINATION = '.jsGiven-reports';

function scenario(options, scenarioFunction) {
    var scenarioOptions = _extends({
        tags: []
    }, options);
    return {
        scenarioFunction,
        tags: scenarioOptions.tags
    };
}

var ScenarioRunner = function () {
    function ScenarioRunner() {
        var reportsDestination = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : REPORTS_DESTINATION;
        classCallCheck(this, ScenarioRunner);

        this.reportsDestination = reportsDestination;
    }

    createClass(ScenarioRunner, [{
        key: 'setup',
        value: function setup(groupFunc, testFunc) {
            this.groupFunc = groupFunc;
            this.testFunc = testFunc;
        }
    }, {
        key: 'scenarios',
        value: function scenarios(groupName, stagesParams, scenariosDescriptions) {
            var _this = this;

            var report = new GroupReport(groupName);

            var currentStages = undefined;

            var stageBuilder = void 0;

            if (Array.isArray(stagesParams)) {
                var self = this;

                var _stagesParams = slicedToArray(stagesParams, 3),
                    givenClass = _stagesParams[0],
                    whenClass = _stagesParams[1],
                    thenClass = _stagesParams[2];

                stageBuilder = function stageBuilder(runningScenario) {
                    var givenStage = self.buildObject(givenClass, runningScenario);
                    var whenStage = self.buildObject(whenClass, runningScenario);
                    var thenStage = self.buildObject(thenClass, runningScenario);
                    return { givenStage, whenStage, thenStage };
                };
            } else {
                var _self = this;
                var _givenClass = stagesParams;

                stageBuilder = function stageBuilder(runningScenario) {
                    var givenStage = _self.buildObject(_givenClass, runningScenario);
                    var whenStage = givenStage;
                    var thenStage = givenStage;
                    return { givenStage, whenStage, thenStage };
                };
            }

            var scenariosParam = {
                given: function given() {
                    if (currentStages) {
                        return currentStages.givenStage.given();
                    } else {
                        throw new Error('given() may only be called in scenario');
                    }
                },
                when: function when() {
                    if (currentStages) {
                        return currentStages.whenStage.when();
                    } else {
                        throw new Error('when() may only be called in scenario');
                    }
                },
                then: function then() {
                    if (currentStages) {
                        return currentStages.thenStage.then();
                    } else {
                        throw new Error('then() may only be called in scenario');
                    }
                }
            };

            this.groupFunc(groupName, function () {
                var scenarios = scenariosDescriptions(scenariosParam);

                getScenarios(scenarios).forEach(function (_ref) {
                    var scenarioPropertyName = _ref.scenarioPropertyName,
                        cases = _ref.cases,
                        argumentNames = _ref.argumentNames;

                    var scenarioNameForHumans = humanize(scenarioPropertyName);
                    var scenario = _this.addScenario(report, scenarioNameForHumans, argumentNames);

                    var casesCount = 0;
                    cases.forEach(function (_ref2, index) {
                        var caseFunction = _ref2.caseFunction,
                            args = _ref2.args;

                        var caseDescription = cases.length === 1 ? scenarioNameForHumans : `${scenarioNameForHumans} #${index + 1}`;
                        _this.testFunc(caseDescription, function () {
                            var runningScenario = {
                                state: 'COLLECTING_STEPS',
                                stages: [],
                                steps: []
                            };

                            _this.addCase(scenario, args);

                            // Build stages
                            currentStages = stageBuilder(runningScenario);

                            // Collecting steps
                            caseFunction();

                            // Execute scenario
                            runningScenario.state = 'RUNNING';
                            var runningSynchronously = true;
                            try {
                                var _ret = function () {
                                    var steps = runningScenario.steps;

                                    var _loop = function _loop(i) {
                                        var _steps$i = steps[i],
                                            stepAction = _steps$i.stepAction,
                                            stage = _steps$i.stage;

                                        var asyncActions = executeStepAndCollectAsyncActions(stepAction);
                                        copyStateToOtherStages(stage, runningScenario.stages);

                                        if (asyncActions.length > 0) {
                                            // If we have asynchronous actions, we switch to asynchronous mode
                                            // We return immediately a promise and the execution is now asynchronous
                                            // This ensures that scenarios that do not require asynchronous processing are still executed synchronously
                                            runningSynchronously = false;
                                            return {
                                                v: {
                                                    v: asyncToGenerator(function* () {
                                                        var executeAsyncActions = function () {
                                                            var _ref4 = asyncToGenerator(function* (asyncActions) {
                                                                var _iteratorNormalCompletion = true;
                                                                var _didIteratorError = false;
                                                                var _iteratorError = undefined;

                                                                try {
                                                                    for (var _iterator = asyncActions[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                                                                        var asyncAction = _step.value;

                                                                        yield asyncAction();
                                                                    }
                                                                } catch (err) {
                                                                    _didIteratorError = true;
                                                                    _iteratorError = err;
                                                                } finally {
                                                                    try {
                                                                        if (!_iteratorNormalCompletion && _iterator.return) {
                                                                            _iterator.return();
                                                                        }
                                                                    } finally {
                                                                        if (_didIteratorError) {
                                                                            throw _iteratorError;
                                                                        }
                                                                    }
                                                                }
                                                            });

                                                            return function executeAsyncActions(_x2) {
                                                                return _ref4.apply(this, arguments);
                                                            };
                                                        }();

                                                        try {
                                                            // Execute async actions for current step
                                                            yield executeAsyncActions(asyncActions);
                                                            copyStateToOtherStages(stage, runningScenario.stages);

                                                            // Execute further steps and their async actions
                                                            for (var j = i + 1; j < steps.length; j++) {
                                                                var _steps$j = steps[j],
                                                                    _stepAction = _steps$j.stepAction,
                                                                    _stage = _steps$j.stage;

                                                                var actions = executeStepAndCollectAsyncActions(_stepAction);
                                                                yield executeAsyncActions(actions);
                                                                copyStateToOtherStages(_stage, runningScenario.stages);
                                                            }
                                                        } finally {
                                                            cleanUp(_this);
                                                        }
                                                    })()
                                                }
                                            };
                                        }
                                    };

                                    for (var i = 0; i < steps.length; i++) {
                                        var _ret2 = _loop(i);

                                        if (typeof _ret2 === "object") return _ret2.v;
                                    }
                                }();

                                if (typeof _ret === "object") return _ret.v;
                            } finally {
                                if (runningSynchronously) {
                                    cleanUp(_this);
                                }
                            }

                            function cleanUp(self) {
                                casesCount++;
                                if (casesCount === cases.length) {
                                    scenario.dumpToFile(self.reportsDestination);
                                }
                                currentStages = undefined;
                            }
                        });
                    });
                });

                function getScenarios(scenarios) {
                    var scenarioDescriptions = Object.keys(scenarios).map(function (scenarioPropertyName) {
                        var scenarioDescription = scenarios[scenarioPropertyName];
                        var scenarioFunction = scenarioDescription.scenarioFunction;

                        if (scenarioFunction instanceof Function) {
                            return {
                                scenarioPropertyName,
                                cases: [{
                                    caseFunction: scenarioFunction,
                                    args: []
                                }],
                                argumentNames: []
                            };
                        } else {
                            var _ref5 = scenarioFunction,
                                _parameters = _ref5.parameters,
                                _func = _ref5.func;

                            var _argumentNames = retrieveArguments(_func);

                            return {
                                scenarioPropertyName,
                                cases: _parameters.map(function (parametersForCase) {
                                    var parametersForTestFunction = parametersForCase.map(function (parameter, index) {
                                        return wrapParameter(parameter, _argumentNames[index]);
                                    });
                                    var args = parametersForCase.map(function (p) {
                                        return formatParameter(p, []);
                                    });
                                    return {
                                        caseFunction: function caseFunction() {
                                            return _func.apply(undefined, toConsumableArray(parametersForTestFunction));
                                        },
                                        args
                                    };
                                }),
                                argumentNames: _argumentNames
                            };
                        }
                    });
                    return scenarioDescriptions;
                }
            });
        }
    }, {
        key: 'addScenario',
        value: function addScenario(report, scenarioNameForHumans, argumentNames) {
            return new ScenarioReport(report, scenarioNameForHumans, [], argumentNames);
        }
    }, {
        key: 'addCase',
        value: function addCase(scenario, args) {
            this.currentCase = new ScenarioCase(args);
            scenario.cases.push(this.currentCase);
        }
    }, {
        key: 'addGivenPart',
        value: function addGivenPart() {
            this.currentPart = new ScenarioPart('GIVEN');
            this.currentCase.parts.push(this.currentPart);
        }
    }, {
        key: 'addWhenPart',
        value: function addWhenPart() {
            this.currentPart = new ScenarioPart('WHEN');
            this.currentCase.parts.push(this.currentPart);
        }
    }, {
        key: 'addThenPart',
        value: function addThenPart() {
            this.currentPart = new ScenarioPart('THEN');
            this.currentCase.parts.push(this.currentPart);
        }
    }, {
        key: 'buildObject',
        value: function buildObject(tClass, runningScenario) {
            var _this3 = this;

            // $FlowIgnore
            var extendedClass = function (_tClass) {
                inherits(extendedClass, _tClass);

                function extendedClass() {
                    classCallCheck(this, extendedClass);
                    return possibleConstructorReturn(this, (extendedClass.__proto__ || Object.getPrototypeOf(extendedClass)).apply(this, arguments));
                }

                return extendedClass;
            }(tClass);

            // Flowtype really can't type this constructor invocation
            // Therefore we have to cast it as any :(


            var instance = new extendedClass();

            var extendedPrototype = Object.getPrototypeOf(instance);
            var classPrototype = Object.getPrototypeOf(extendedPrototype);

            var stages = runningScenario.stages;


            getAllMethods(classPrototype).forEach(function (methodName) {
                var self = _this3;

                extendedPrototype[methodName] = function () {
                    var _this4 = this;

                    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
                        args[_key] = arguments[_key];
                    }

                    var state = runningScenario.state,
                        steps = runningScenario.steps;

                    switch (state) {
                        case 'COLLECTING_STEPS':
                            {
                                steps.push({
                                    stepAction: function stepAction() {
                                        return extendedPrototype[methodName].apply(_this4, args);
                                    },
                                    stage: this
                                });
                                return this;
                            }
                        default:
                            {
                                // eslint-disable-next-line no-unused-vars
                                var typeCheck = state;

                                if (methodName === 'given') {
                                    self.addGivenPart();
                                }
                                if (methodName === 'when') {
                                    self.addWhenPart();
                                }
                                if (methodName === 'then') {
                                    self.addThenPart();
                                }

                                var stepParameterNames = retrieveArguments(classPrototype[methodName]);

                                var decodedParameters = args.map(function (arg, index) {
                                    if (index < stepParameterNames.length) {
                                        return decodeParameter(arg, stepParameterNames[index], getFormatters(instance, methodName, stepParameterNames[index]));
                                    } else {
                                        return decodeParameter(arg, restParameterName(), getFormatters(instance, methodName, restParameterName()));
                                    }
                                });

                                // Pass the real arguments instead of the wrapped values
                                var values = decodedParameters.map(function (decodedParameter) {
                                    return decodedParameter.value;
                                });
                                var result = classPrototype[methodName].apply(this, values);

                                if (result === this && !isHiddenStep(this, methodName)) {
                                    // only records methods that return this
                                    self.currentPart.stageMethodCalled(methodName, decodedParameters);
                                }
                                return result;
                            }
                    }
                };
            });

            stages.push(instance);

            return instance;

            function getAllMethods(obj) {
                var allMethods = [];
                var current = obj;
                do {
                    var props = Object.getOwnPropertyNames(current);
                    props.forEach(function (prop) {
                        if (allMethods.indexOf(prop) === -1) {
                            if (_.isFunction(current[prop])) {
                                allMethods.push(prop);
                            }
                        }
                    });
                } while (current = Object.getPrototypeOf(current));

                return allMethods;
            }
        }
    }]);
    return ScenarioRunner;
}();

var INSTANCE = new ScenarioRunner();

function scenarios(groupName, stagesParam, scenarioFunc) {
    return INSTANCE.scenarios(groupName, stagesParam, scenarioFunc);
}

function copyStateToOtherStages(originalStage, allStages) {
    allStages.forEach(function (targetStage) {
        if (originalStage !== targetStage) {
            copyStateProperties(originalStage, targetStage);
        }
    });
}

function parametrized(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}

function parametrized1(parameters, func) {
    return {
        parameters: parameters.map(function (param) {
            return [param];
        }),
        func
    };
}
function parametrized2(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized3(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized4(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized5(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized6(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}
function parametrized7(parameters, func) {
    return {
        parameters: parameters,
        func
    };
}

function wrapParameter(value, scenarioParameterName) {
    return {
        scenarioParameterName,
        value,
        IS_JSGIVEN_WRAPPER_PARAMETER: true
    };
}

function decodeParameter(parameter, stepParameterName, formatters) {
    if (parameter instanceof Object && parameter.IS_JSGIVEN_WRAPPER_PARAMETER) {
        var wrapped = parameter;
        return _extends({}, wrapped, { stepParameterName, formatters });
    } else {
        if (stepParameterName === undefined) {
            throw new Error('cant be undefined');
        }
        return {
            value: parameter,
            scenarioParameterName: null,
            stepParameterName,
            formatters
        };
    }
}

var asyncActionsSingleton = [];
function executeStepAndCollectAsyncActions(stepActionThatMayCallDoAsync) {
    asyncActionsSingleton = [];
    var collectedAsyncActions = [];

    try {
        stepActionThatMayCallDoAsync();
    } finally {
        collectedAsyncActions.push.apply(collectedAsyncActions, toConsumableArray(asyncActionsSingleton));
        asyncActionsSingleton = [];
    }

    return collectedAsyncActions;
}

function doAsync(action) {
    asyncActionsSingleton.push(action);
}

function setupForRspec(describe, it) {
    return INSTANCE.setup(function (groupName, suiteFunc) {
        describe(groupName, suiteFunc);
    }, function (testName, testFunc) {
        it(testName, testFunc);
    });
}

function setupForAva(test) {
    var capturedGroupName = '';
    return INSTANCE.setup(function (groupName, suiteFunc) {
        capturedGroupName = groupName;
        suiteFunc();
    }, function (testName, testFunc) {
        test(`${capturedGroupName} / ${testName}`, testFunc);
    });
}

exports.parametrized = parametrized;
exports.parametrized1 = parametrized1;
exports.parametrized2 = parametrized2;
exports.parametrized3 = parametrized3;
exports.parametrized4 = parametrized4;
exports.parametrized5 = parametrized5;
exports.parametrized6 = parametrized6;
exports.parametrized7 = parametrized7;
exports.scenario = scenario;
exports.scenarios = scenarios;
exports.doAsync = doAsync;
exports.setupForRspec = setupForRspec;
exports.setupForAva = setupForAva;
exports.Stage = Stage;
exports.State = State;
exports.Hidden = Hidden;
exports.Quoted = Quoted;
exports.QuotedWith = QuotedWith;
exports.NotFormatter = NotFormatter;
exports.buildParameterFormatter = buildParameterFormatter;
